"""

        This is SlideRunner - An Open Source Annotation Tool 
        for Digital Histology Slides.

         Marc Aubreville, Pattern Recognition Lab, 
         Friedrich-Alexander University Erlangen-Nuremberg 
         marc.aubreville@fau.de

        If you use this software in research, please citer our paper:
        M. Aubreville, C. Bertram, R. Klopfleisch and A. Maier:
        SlideRunner - A Tool for Massive Cell Annotations in Whole Slide Images
        Bildverarbeitung fuer die Medizin 2018, Springer Verlag, Berlin-Heidelberg

"""

__all__ = ['database', 'slide', 'dicomimage', 'annotations', 'nifty']

from .slide import os_fileformats
from .nifty import nii_fileformats
from .dicomimage import dic_fileformats

fileformats = os_fileformats + nii_fileformats + dic_fileformats

